#
# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

import argparse

from .connections import MDDBConnector
from .queries import (
    create_comments,
    delete_comments,
    get_all_spls_from_exec_blocks,
    get_exec_blocks,
    modify_comments,
    retrieve_problem_severities,
    retrieve_problem_types,
    retrieve_acknowledgements,
    create_acknowledgement,
    modify_acknowledgement,
    delete_acknowledgement,
    assign_acknowledgement_to_project,
    delete_acknowledgement_from_project
)


def parser() -> argparse.ArgumentParser:
    """
    Initialize argument parser for command-line arguments

    :return: Argument parser
    """
    parser = argparse.ArgumentParser(
        description="Workspaces data annotator", formatter_class=argparse.RawTextHelpFormatter
    )

    # create subparsers for each mode of
    subparsers = parser.add_subparsers(dest="subcommand")
    create_data_notes_subparser(subparsers)
    create_proj_acks_subparser(subparsers)
    create_ack_text_subparser(subparsers)

    return parser


def create_data_notes_subparser(subparsers):
    data_notes_subparser = subparsers.add_parser('notes', help='Create, modify, or delete a data comment for a dataset.')

    # Create a group to capture the action to be done; require that users select exactly one of this group
    action_group = data_notes_subparser.add_mutually_exclusive_group(required=True)
    action_group.add_argument(
        "-c",
        "--create",
        action="store_true",
        help="Create a comment for one or more datasets. Cannot be used with -d/--delete or -m/--modify.",
    )
    action_group.add_argument(
        "-m",
        "--modify",
        action="store_true",
        help="Modify a comment for one or more datasets. Cannot be used with -d/--delete or -c/--create.",
    )
    action_group.add_argument(
        "-d",
        "--delete",
        action="store_true",
        help="Delete a comment for one or more datasets. Cannot be used with -c/--create or -m/--modify.",
    )

    # Create a group to capture the data that will be searched by when querying the database; require at least one
    search_group = data_notes_subparser.add_argument_group()
    search_group.add_argument(
        "--project-code",
        action="store",
        help="A string representing the project code to search by.",
    )
    search_group.add_argument(
        "--date-range",
        nargs=2,
        action="store",
        help="Two strings representing beginning and end date to search by.",
    )
    search_group.add_argument(
        "--external-name",
        action="store",
        help="A string representing the external name to search by.",
    )
    search_group.add_argument(
        "--configuration",
        action="store",
        help="A string representing the configuration to search by.",
    )
    search_group.add_argument(
        "--band",
        action="store",
        help="A string representing the band to search by.",
    )

    # Create a group to capture other data that is not required at the start of the program
    optional_group = data_notes_subparser.add_argument_group()
    optional_group.add_argument(
        "--comment",
        action="store",
        help="A string representing the comment to add.",
    )
    optional_group.add_argument(
        "--old-comment",
        action="store",
        help="A string representing the old comment to search by. Only used with -m/--modify.",
    )
    optional_group.add_argument(
        "--problem-severity",
        action="store",
        help="An integer between 0-3 with 0 being not a problem, and 1-3 being low, medium and high severity.",
    )
    optional_group.add_argument(
        "--problem-type",
        action="store",
        help="An integer from 1-10; 1. Comment/Other, 2. Clock Error, 3. Position Error, 4. Model Error, 5. Observing Error, 6. Hardware Error, 7. Session ID, 8. Data Errors, 9. Missing Data, 10. Processing Error.",
    )


def create_proj_acks_subparser(subparsers):
    acknowledgements_subparser = subparsers.add_parser('project-acks', help='Assign or remove acknowledgements for a project.')

    # Create a group to capture the action to be done; require that users select exactly one of this group
    action_group = acknowledgements_subparser.add_mutually_exclusive_group(required=True)
    action_group.add_argument(
        "-a",
        "--assign",
        action="store_true",
        help="Add an acknowledgement to a project. Cannot be used with -d/--delete.",
    )
    action_group.add_argument(
        "-d",
        "--delete",
        action="store_true",
        help="Delete an acknowledgement from a project. Cannot be used with -a/--assign.",
    )

    # Create a group to capture the data that will be searched by when querying the database; require at least one
    search_group = acknowledgements_subparser.add_argument_group()
    search_group.add_argument(
        "--project-code",
        action="store",
        help="A string representing the project code to search by.",
    )
    search_group.add_argument(
        "--acknowledgement-id",
        action="store",
        help="An integer representing the acknowledgement to use.",
    )


def create_ack_text_subparser(subparsers):
    """
    Adds a parser meant to handle modification of acknowledgment text.

    :param subparsers: a subparser set that requires these flags
    """
    acknowledgements_subparser = subparsers.add_parser('ack-text', help='Add, modify, or delete acknowledgements.')

    # Create a group to capture the action to be done; require that users select exactly one of this group
    action_group = acknowledgements_subparser.add_mutually_exclusive_group(required=True)
    action_group.add_argument(
        "-c",
        "--create",
        action="store_true",
        help="Create an acknowledgement. Cannot be used with -d/--delete or -m/--modify.",
    )
    action_group.add_argument(
        "-m",
        "--modify",
        action="store_true",
        help="Modify an acknowledgement. Cannot be used with -d/--delete or -c/--create.",
    )
    action_group.add_argument(
        "-d",
        "--delete",
        action="store_true",
        help="Delete an acknowledgement. Cannot be used with -c/--create or -m/--modify.",
    )

    # Create a group to capture the data that will be searched by when querying the database; require at least one
    search_group = acknowledgements_subparser.add_argument_group()
    search_group.add_argument(
        "--institution",
        action="store",
        help="A string representing the institution (e.g. ALMA, NRAO) to associate with an acknowledgement.",
    )
    search_group.add_argument(
        "--acknowledgement-id",
        action="store",
        help="A string representing the acknowledgement id to use.",
    )

    # Create a group to capture other data that is not required at the start of the program
    optional_group = acknowledgements_subparser.add_argument_group()
    optional_group.add_argument(
        "--text",
        action="store",
        help="A string representing the new acknowledgement text to assign or change to.",
    )


def prompt_for_input_from_db_list(db_list, prompt):
    """
    Accepts a list of id and description pairs and a sentence prompt to give to a user, and returns the user's choice
    from this list.

    :param db_list: a list of id and description tuples
    :param prompt: a string representing a question to ask the user
    :return: the user's choice from the passed iin options
    """
    output_string = prompt
    for item in db_list:
        output_string += str(item[0]) + ".  " + item[1]
        output_string += f".  {str(item[2])}" if len(item) > 2 else ""
        output_string += "\n"

    return input(output_string)


def prompt_for_missing_notes_info(args, cursor):
    """
    Checks for required information for a data note change and retrieves any missing data.

    :param args: an object holding all current user arguments
    :param cursor: an object representing a database connection
    """
    if not args.problem_severity:
        problem_severities = retrieve_problem_severities(cursor)
        prompt = "A problem severity is required. Input the number for the value you wish to assign and hit enter:\n"
        args.problem_severity = prompt_for_input_from_db_list(problem_severities, prompt)
    if not args.problem_type:
        problem_types = retrieve_problem_types(cursor)
        prompt = "A problem type is required. Input the number for the value you wish to assign and hit enter:\n"
        args.problem_type = prompt_for_input_from_db_list(problem_types, prompt)
    if not args.comment:
        prompt = "Please type your comment and hit enter:\n"
        args.comment = input(prompt)


def prompt_for_missing_project_ack_info(args, cursor):
    """
    Checks for required information for a project acknowledgement change and retrieves any missing data.

    :param args: an object holding all current user arguments
    :param cursor: an object representing a database connection
    """
    if not args.project_code:
        prompt = "A project code is required. Input the code for the project you wish to assign and hit enter:\n"
        args.project_code = input(prompt)
    if not args.acknowledgement_id:
        acknowledgements = retrieve_acknowledgements(cursor)
        prompt = "An acknowledgement id is required. Input the number for the value you wish to assign and hit enter:\n"
        args.acknowledgement_id = prompt_for_input_from_db_list(acknowledgements, prompt)


def prompt_for_missing_ack_text_info(args, cursor):
    """
    Checks for required information for an acknowledgement text change and retrieves any missing data.

    :param args: an object holding all current user arguments
    :param cursor: an object representing a database connection cursor
    """
    if not args.acknowledgement_id and not args.create:
        acknowledgements = retrieve_acknowledgements(cursor)
        prompt = "An acknowledgement id is required. Input the number for the value you wish to assign and hit enter:\n"
        args.acknowledgement_id = prompt_for_input_from_db_list(acknowledgements, prompt)
    if not args.text and not args.delete:
        prompt = "New text for the acknowledgement is required. Input the text you wish to assign and hit enter:\n"
        args.text = input(prompt)
    if not args.institution and args.create:
        prompt = "An institution for the acknowledgement is required. Input the institution you wish to assign and hit enter:\n"
        args.institution = input(prompt)


def main():
    print("**********************************")
    print("* WELCOME TO THE DATA ANNOTATOR! *")
    print("**********************************")

    args = parser().parse_args()
    try:
        conn = MDDBConnector()
    except Exception as e:
        print(e)

    cursor = conn.cursor()
    if args.subcommand == 'project-acks':
        prompt_for_missing_project_ack_info(args, cursor)
        if args.assign:
            assign_acknowledgement_to_project(args.acknowledgement_id, args.project_code, cursor)
        elif args.delete:
            delete_acknowledgement_from_project(args.acknowledgement_id, args.project_code, cursor)

    elif args.subcommand == 'ack-text':
        prompt_for_missing_ack_text_info(args, cursor)
        if args.create:
            create_acknowledgement(args.institution, args.text, cursor)
        elif args.delete:
            delete_acknowledgement(args.acknowledgement_id, cursor)
        elif args.modify:
            modify_acknowledgement(args.acknowledgement_id, args.text, cursor)

    elif args.subcommand == 'notes':
        prompt_for_missing_notes_info(args, cursor)

        # Find the initial execution blocks, use them to find all relevant science product locators and proceed
        ebs = get_exec_blocks(args.project_code, args.date_range, args.external_name, args.band, args.configuration, cursor)
        all_spls = get_all_spls_from_exec_blocks(ebs, cursor)
        if args.create:
            create_comments(all_spls, args.comment, args.problem_severity, args.problem_type, cursor)
        elif args.delete:
            delete_comments(all_spls, args.comment, args.problem_severity, args.problem_type, cursor)
        elif args.modify:
            if not args.old_comment:
                prompt = "Please type the old comment you wish to modify:\n"
                args.old_comment = input(prompt)
            modify_comments(all_spls, args.comment, args.old_comment, args.problem_severity, args.problem_type, cursor)

    conn.commit()
    conn.close()


if __name__ == "__main__":
    main()
