#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import pathlib
import shutil
from test import (
    RESTORE_AUXPRODUCTS_NAME,
    TEST_RESTORE_METADATA,
    get_expected_files_and_dirs_for_restore,
)
from test.conftest import assert_restore_delivered_only_expected_files
from typing import Dict

from delivery.destinations.interfaces import Destination, DestinationTempFile
from delivery.finder import RestoreProductFinder
from delivery.products import Calibration, ProductMetadata, RestoreProduct


class FakeDestination(Destination):
    def __init__(self):
        super().__init__(None)
        self.files_added: dict[str, pathlib.Path] = {}

    def add_file(self, file: pathlib.Path, relative_path: str):
        self.files_added[relative_path] = file

    def create_file(self, relative_path: str) -> DestinationTempFile:
        raise NotImplementedError

    def results(self) -> Dict:
        return {}


def test_web_calibration(resource_path_root):
    calibration = Calibration(resource_path_root / "calibrations" / "foo")

    # first, let's make sure we parsed the metadata properly
    project_code, pipeline_id = calibration.parse_metadata()
    assert project_code == "21A-409"
    assert pipeline_id == "59311.953874212966"

    # OK now let's make sure that we're delivering properly
    destination = FakeDestination()
    calibration.deliver_to(destination)

    expected_files = set(
        [
            "21A-409/calibration_pipeline.59311.953874212966/pipeline-20210409T164741/.keep",
            "21A-409/calibration_pipeline.59311.953874212966/unknown.hifv_cal.pipeline_manifest.xml",
            "21A-409/calibration_pipeline.59311.953874212966/casa_commands.log",
            "21A-409/calibration_pipeline.59311.953874212966/flux.csv",
            "21A-409/calibration_pipeline.59311.953874212966/unknown.hifv_cal.pprequest.xml",
            "21A-409/calibration_pipeline.59311.953874212966/21A-409.sb39530507.eb39564922.59311.953874212966.ms.calapply.txt",
            "21A-409/calibration_pipeline.59311.953874212966/casa_pipescript.py",
            "21A-409/calibration_pipeline.59311.953874212966/casa_piperestorescript.py",
            "21A-409/calibration_pipeline.59311.953874212966/PPR.xml",
            "21A-409/calibration_pipeline.59311.953874212966/pipeline_aquareport.xml",
            "21A-409/calibration_pipeline.59311.953874212966/weblog.tgz",
        ]
    )
    actual_files = destination.files_added.keys()

    # did we get all the expected files?
    for file in expected_files:
        assert file in actual_files

    # did we not get any unexpected files?
    for file in actual_files:
        assert file in expected_files


def test_restore(restore_directory: pathlib.Path):
    restore = RestoreProduct(restore_directory, TEST_RESTORE_METADATA)

    # OK now let's make sure that we're delivering properly
    destination = FakeDestination()
    restore.deliver_to(destination)

    deliver_rel_path_root = f"{TEST_RESTORE_METADATA.project_code}/{TEST_RESTORE_METADATA.pipeline_spec}"
    # Just check the file counts in each dir since they contain a lot of files
    actual_files = list(pathlib.Path(f) for f in destination.files_added.keys())
    files_to_is_file = {pathlib.Path(key): value.is_file() for key, value in destination.files_added.items()}
    expected_files, expected_dirs_to_fcs = get_expected_files_and_dirs_for_restore(deliver_rel_path_root)
    assert_restore_delivered_only_expected_files(actual_files, expected_files, expected_dirs_to_fcs, files_to_is_file)


def test_restore_with_flagtemplate(restore_directory: pathlib.Path, tmp_path: pathlib.Path):
    auxproducts_path = restore_directory.parent / RESTORE_AUXPRODUCTS_NAME
    assert auxproducts_path.is_file()
    # To avoid possibly polluting the repo,
    #     use a tmp dir for the restore since the auxproducts tar needs to be copied into it
    shutil.copytree(restore_directory, tmp_path, dirs_exist_ok=True)
    shutil.copy2(auxproducts_path, tmp_path / "products")

    restore = RestoreProduct(tmp_path, TEST_RESTORE_METADATA)

    # OK now let's make sure that we're delivering properly
    destination = FakeDestination()
    restore.deliver_to(destination)

    deliver_rel_path_root = f"{TEST_RESTORE_METADATA.project_code}/{TEST_RESTORE_METADATA.pipeline_spec}"
    # Just check the file counts in each dir since they contain a lot of files
    actual_files = list(pathlib.Path(f) for f in destination.files_added.keys())
    files_to_is_file = {pathlib.Path(key): value.is_file() for key, value in destination.files_added.items()}
    expected_files, expected_dirs_to_fcs = get_expected_files_and_dirs_for_restore(
        deliver_rel_path_root, do_include_flagtemplate=True
    )
    assert_restore_delivered_only_expected_files(actual_files, expected_files, expected_dirs_to_fcs, files_to_is_file)


def test_restore_product_finder_parse_metadata(restore_directory: pathlib.Path):
    expected_metadata = TEST_RESTORE_METADATA
    finder = RestoreProductFinder(restore_directory)
    actual_metadata = finder.parse_metadata()
    assert actual_metadata == expected_metadata
